unit MultiUpdateSQL;
{
Author: Craig Manley
Date: 10 January 2000
Version: 1.1

Legal issues:
=============
  Copyright (C) 2000 by Craig Manley.
  Email: craig@skybound.nl
  Website: http://www.skybound.nl

  This software is provided 'as-is', without any express or
  implied warranty. In no event will the author be held liable
  for any damages arising from the use of this software.

  Permission is granted to anyone to use this software for any
  purpose, including commercial applications, and to alter it
  and redistribute it freely, subject to the following
  restrictions:

  1. The origin of this software must not be misrepresented,
     you must not claim that you wrote the original software.
     If you use this software in a product, an acknowledgment
     in the product documentation would be appreciated but is
     not required.

  2. Altered source versions must be plainly marked as such, and
     must not be misrepresented as being the original software.

  3. This notice may not be removed or altered from any source
     distribution.
}


interface

uses
  Windows, Classes, SysUtils, DB, DBTables, ListStuff;

type
  PStatement = ^TStatement;
  TStatement = record
    SQL: string;
  end;

  TExecuteEvent = procedure(Sender: TObject; Index: integer; const Statement: string) of object;

  TCheckRowsAffected = (craAlways,craLastStatementOnly,craNever);

  TMultiUpdateSQL = class(TDataSetUpdateObject)
  private
    FSQL: TContainerList;
    FCheckRowsAffected: TCheckRowsAffected;
    FDataSet: TBDEDataSet;
    FOnAfterExecute: TExecuteEvent;
    FOnBeforeExecute: TExecuteEvent;
    FQueries: array[TUpdateKind] of TQuery;
    FSQLText: array[TUpdateKind] of TStrings;
    FTerminator: string;
    function GetCount(UpdateKind: TUpdateKind): Cardinal;
    function GetQuery(UpdateKind: TUpdateKind): TQuery;
    function GetSQL(UpdateKind: TUpdateKind): TStrings;
    function GetSQLIndex(Index: Integer): TStrings;
    procedure SetSQL(UpdateKind: TUpdateKind; Value: TStrings);
    procedure SetSQLIndex(Index: Integer; Value: TStrings);
    function GetStatement(UpdateKind: TUpdateKind; Index: integer): string;
    procedure SetTerminator(const AValue: string);
  protected
    procedure SetParams(UpdateKind: TUpdateKind);
    function GetDataSet: TBDEDataSet; override;
    procedure SetDataSet(ADataSet: TBDEDataSet); override;
    procedure SQLChanged(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Apply(UpdateKind: TUpdateKind); override;
    property Count[UpdateKind: TUpdateKind]: Cardinal read GetCount;
    property DataSet;
    property Query[UpdateKind: TUpdateKind]: TQuery read GetQuery;
    property SQL[UpdateKind: TUpdateKind]: TStrings read GetSQL write SetSQL;
    property Statements[UpdateKind: TUpdateKind; Index: integer]: string read GetStatement;
  published
    property ModifySQL: TStrings index 0 read GetSQLIndex write SetSQLIndex;
    property InsertSQL: TStrings index 1 read GetSQLIndex write SetSQLIndex;
    property DeleteSQL: TStrings index 2 read GetSQLIndex write SetSQLIndex;
    property CheckRowsAffected: TCheckRowsAffected read FCheckRowsAffected write FCheckRowsAffected;
    property Terminator: string read FTerminator write SetTerminator;
    property OnAfterExecute: TExecuteEvent read FOnAfterExecute write FOnAfterExecute;
    property OnBeforeExecute: TExecuteEvent read FOnBeforeExecute write FOnBeforeExecute;
  end;

procedure Register;

implementation

uses BDEConst, StrStuff;

{$R *.RES}

procedure Register;
begin
  RegisterComponents('Freeware', [TMultiUpdateSQL]);
end;


{TMultiUpdateSQL}
procedure SplitStatements(ListOfStatements: TContainerList; Statements: TStrings; const Terminator: string);
var
  i: integer;
  s: string;
  Statement: string;
  p: PStatement;
begin
  ListOfStatements.Clear;
  Statement := '';
  for i := 0 to Statements.Count-1 do
  begin
    s := TrimRight(Statements.Strings[i]);
    if (Length(s) > 0) and (LastPos(Terminator,s) = 1 + Length(s) - Length(Terminator)) then
    begin
      Statement := Statement + Copy(s,1,Length(s)-Length(Terminator));
      new(p);
      p^.SQL := Statement;
      ListOfStatements.Add(p);
      Statement := '';
    end
    else
    begin
      Statement := Statement + Statements.Strings[i] + #13#10;
      if (i = Statements.Count-1) then
      begin
        new(p);
        p^.SQL := Statement;
        ListOfStatements.Add(p);
        Statement := '';
      end;
    end;
  end;
end;


constructor TMultiUpdateSQL.Create(AOwner: TComponent);
var
  UpdateKind: TUpdateKind;
begin
  inherited Create(AOwner);
  FSQL := TContainerList.Create(rtStruct);
  for UpdateKind := Low(TUpdateKind) to High(TUpdateKind) do
  begin
    FSQLText[UpdateKind] := TStringList.Create;
    TStringList(FSQLText[UpdateKind]).OnChange := SQLChanged;
  end;
  FCheckRowsAffected := craAlways;
  FTerminator := ';';
end;

destructor TMultiUpdateSQL.Destroy;
var
  UpdateKind: TUpdateKind;
begin
  if Assigned(FDataSet) and (FDataSet.UpdateObject = Self) then
    FDataSet.UpdateObject := nil;
  for UpdateKind := Low(TUpdateKind) to High(TUpdateKind) do
    FSQLText[UpdateKind].Free;
  FSQL.Free;
  inherited Destroy;
end;

procedure TMultiUpdateSQL.Apply(UpdateKind: TUpdateKind);
var
  i: integer;
begin
  with Query[UpdateKind] do
  begin
    SplitStatements(FSQL,FSQLText[UpdateKind],FTerminator);
    for i := 0 to FSQL.Count-1 do
    begin
      Params.Clear;
      SQL.Text := TStatement(FSQL.Items[i]^).SQL;
      if Assigned(FOnBeforeExecute) then
        FOnBeforeExecute(self,i,TStatement(FSQL.Items[i]^).SQL);
      SetParams(UpdateKind);
      Prepare;
      ExecSQL;
      if ((CheckRowsAffected = craAlways) or ((CheckRowsAffected = craLastStatementOnly) and (i = FSQL.Count-1))) and (RowsAffected <> 1) then
        DatabaseError(SUpdateFailed);
      if Assigned(FOnAfterExecute) then
        FOnAfterExecute(self,i,TStatement(FSQL.Items[i]^).SQL);
    end;
    FSQL.Clear;
  end;
end;

function TMultiUpdateSQL.GetQuery(UpdateKind: TUpdateKind): TQuery;
begin
  if not Assigned(FQueries[UpdateKind]) then
  begin
    FQueries[UpdateKind] := TQuery.Create(Self);
    //FQueries[UpdateKind].SQL.Assign(FSQLText[UpdateKind]);
    if (FDataSet is TDBDataSet) then
    begin
      FQueries[UpdateKind].SessionName := TDBDataSet(FDataSet).SessionName;
      FQueries[UpdateKind].DatabaseName := TDBDataSet(FDataSet).DataBaseName;
    end;
  end;
  Result := FQueries[UpdateKind];
end;

function TMultiUpdateSQL.GetSQL(UpdateKind: TUpdateKind): TStrings;
begin
  Result := FSQLText[UpdateKind];
end;

function TMultiUpdateSQL.GetSQLIndex(Index: Integer): TStrings;
begin
  Result := FSQLText[TUpdateKind(Index)];
end;

function TMultiUpdateSQL.GetDataSet: TBDEDataSet;
begin
  Result := FDataSet;
end;

procedure TMultiUpdateSQL.SetDataSet(ADataSet: TBDEDataSet);
begin
  FDataSet := ADataSet;
end;

procedure TMultiUpdateSQL.SetSQL(UpdateKind: TUpdateKind; Value: TStrings);
begin
  FSQLText[UpdateKind].Assign(Value);
end;

procedure TMultiUpdateSQL.SetSQLIndex(Index: Integer; Value: TStrings);
begin
  SetSQL(TUpdateKind(Index), Value);
end;

procedure TMultiUpdateSQL.SQLChanged(Sender: TObject);
var
  UpdateKind: TUpdateKind;
begin
  for UpdateKind := Low(TUpdateKind) to High(TUpdateKind) do
    if Sender = FSQLText[UpdateKind] then
    begin
      if Assigned(FQueries[UpdateKind]) then
      begin
        FQueries[UpdateKind].Params.Clear;
        FQueries[UpdateKind].SQL.Assign(FSQLText[UpdateKind]);
      end;
      Break;
    end;
end;

procedure TMultiUpdateSQL.SetParams(UpdateKind: TUpdateKind);
var
  I: Integer;
  Old: Boolean;
  Param: TParam;
  PName: string;
  Field: TField;
  Value: Variant;
begin
  if not Assigned(FDataSet) then Exit;
  with Query[UpdateKind] do
  begin
    for I := 0 to Params.Count - 1 do
    begin
      Param := Params[I];
      PName := Param.Name;
      Old := CompareText(Copy(PName, 1, 4), 'OLD_') = 0;
      if Old then System.Delete(PName, 1, 4);
      Field := FDataSet.FindField(PName);
      if not Assigned(Field) then Continue;
      if Old then Param.AssignFieldValue(Field, Field.OldValue) else
      begin
        Value := Field.NewValue;
        if VarIsEmpty(Value) then Value := Field.OldValue;
        Param.AssignFieldValue(Field, Value);
      end;
    end;
  end;
end;

procedure TMultiUpdateSQL.SetTerminator(const AValue: string);
begin
  if AValue > '' then
    FTerminator := AValue;
end;

function TMultiUpdateSQL.GetCount(UpdateKind: TUpdateKind): Cardinal;
var
  List: TContainerList;
begin
  List := TContainerList.Create(rtStruct);
  try
    SplitStatements(List, FSQLText[UpdateKind], FTerminator);
    Result := List.Count;
  finally
    List.Free;
  end;
end;

function TMultiUpdateSQL.GetStatement(UpdateKind: TUpdateKind; Index: integer): string;
var
  List: TContainerList;
begin
  List := TContainerList.Create(rtStruct);
  try
    SplitStatements(List, FSQLText[UpdateKind], FTerminator);
    if (Index < 0) or (Index > List.Count-1) then
      Result := ''
    else
      Result := TStatement(List.Items[Index]^).SQL;
  finally
    List.Free;
  end;
end;

end.
